class Diff
  unloadable

  def self.collate_properties_by_resource(node_baseline)
    refs = Hash.new do |h,k|
      h[k] = Hash.new { |h,k| h[k] = { :inspected => nil, :baseline => nil } }
    end

    baseline = node_baseline.baseline.resources.map do |resource|
      key = [resource.resource_type, resource.title]
      resource.properties.each { |p| refs[key][p.name][:baseline] = p.value }
    end

    differences = node_baseline.differences.map do |diff|
      res = (diff.baseline_resource || diff.report_resource)
      key = [res.resource_type, res.title]

      diff.baseline_resource.properties.each do |p|
        refs[key][p.name][:baseline] = p.value
      end if diff.baseline_resource_id

      diff.report_resource.events.each do |p|
        refs[key][p.property][:inspected] = p.previous_value
      end if diff.report_resource_id
    end

    refs.map { |k,v| [ k, v.sort_by(&:first) ] }.sort_by(&:first)
  end

  def self.node_against_baseline(node_id, time, start, count)
    date = time.to_date
    node_baseline = NodeBaseline.self_comparisons.find_by_node_id_and_date(node_id, date)
    self.node_diff(node_baseline, start, count)
  end

  def self.node_diff(node_baseline, start, count, type=nil, title=nil)
    node_baseline.differences.filter_by_resource(type, title).all(
      :include => {:report_resource => :events, :baseline_resource => :properties },
      :order => 'LOWER(resource_name) ASC, baseline_properties.name ASC',
      :offset => start,
      :limit => count
    ).collect {|diff| node_diff_one diff }
  end

  def self.node_diff_one(diff)
    res = (diff.report_resource || diff.baseline_resource)

    baseline_properties = diff.baseline_resource.properties.map {|p| [p.name, p.value]} rescue []
    report_properties   = diff.report_resource.events.map {|ev| [ev.property, ev.previous_value] } rescue []

    properties = Hash.new {|h,k| h[k] = {}}

    baseline_properties.each do |name, value|
      properties[name].merge! :baseline => value
    end

    report_properties.each do |name, value|
      if properties.key?(name) and properties[name][:baseline] == value
        properties.delete(name)
      else
        properties[name].merge! :inspected => value
      end
    end

    {
      :resource_name  => "#{res.resource_type}[#{res.title}]",
      :resource_type  => res.resource_type,
      :resource_title => res.title,
      :properties     => properties,
      :accepted       => diff.accepted,
      :diff_id        => diff.id,
      :can_change     => diff.can_change?
    }
  end

  def self.group_diff(node_baselines, start, count, type=nil, title=nil)
    return [] unless node_baselines and node_baselines.present?
    node_baseline_ids = node_baselines.map(&:id)

    resources_to_diff = BaselineDifference.filter_by_resource(type, title).all(
      :select     => "resource_name, GROUP_CONCAT(DISTINCT
                        COALESCE(resource_status_digest, 'no digest')
                      ) AS resource_status_digest",
      :joins      => :node_baseline,
      :conditions => { :node_baseline_id => node_baseline_ids },
      :group      => "resource_name",
      :having     => 'COUNT(distinct node_baselines.node_id) > 1',
      :offset     => start,
      :limit      => count
    ).map do |baseline|
      baseline.resource_status_digest.split(',').map do |digest|
        [ baseline.resource_name, digest ]
      end
    end.flatten

    return [] unless resources_to_diff.present?

    differences = BaselineDifference.all(
      :conditions => [
        "`baseline_differences`.node_baseline_id IN (?) AND (resource_name, coalesce(resource_status_digest, 'no digest')) IN (#{(1..resources_to_diff.length/2).map{'(?,?)'}.join(',')})",
        node_baseline_ids,
        *resources_to_diff
      ],
      :select => "max(id) AS id,
                  max(report_resource_id) AS report_resource_id,
                  max(baseline_resource_id) AS baseline_resource_id,
                  resource_name, resource_status_digest,
                  count(*) AS node_count,
                  MIN(decided_at) AS decided_at,
                  CASE
                    WHEN min(CASE WHEN accepted = true  THEN 1 ELSE 0 END) = 1 THEN 'accepted'
                    WHEN min(CASE WHEN accepted = false THEN 1 ELSE 0 END) = 1 THEN 'rejected'
                    WHEN min(CASE WHEN accepted IS NULL THEN 1 ELSE 0 END) = 1 THEN 'no-action'
                    ELSE                                                            'multiple'
                  END AS accept_decision",
      :group => "resource_name, COALESCE(baseline_resource_digest, 'no digest'), COALESCE(resource_status_digest, 'no digest')",
      :include => {:report_resource => :events, :baseline_resource => :properties}
    )

    resource_outcomes = differences.group_by(&:resource_name)

    results = []

    resource_outcomes.sort.each do |resource, outcomes|
      results << [ resource, props = [] ]

      outcomes.group_by(&:resource_status_digest).each do |resource_digest, diffs|
        node_count = diffs.map{|d| d.node_count.to_i}.sum
        accepted = diffs.map(&:accept_decision).uniq.length == 1 ? diffs.first.accept_decision : 'multiple'

        baseline_properties = diffs.map do |diff|
          next unless diff.baseline_resource
          diff.baseline_resource.properties.map do |prop|
            { prop.name => prop.value }
          end.inject(:merge)
        end.compact

        inspected_properties = {}
        if resource = diffs.first.report_resource
          full = ! baseline_properties.empty?

          resource.events.each do |event|
            name  = event.property
            value = event.previous_value

            # We need to include this value in the inspected side if it is
            # mismatched with *any* of the baseline values we have here.
            #
            # Note that an empty array returns *true* from the all? call,
            # rather than false, hence the `full` test.
            if full and baseline_properties.all? {|x| x.key?(name) and x[name] == value }
              # ...identical everywhere, remove it from the baseline side too.
              baseline_properties = baseline_properties.map do |this|
                this.reject {|k,v| k == name }
              end
            else
              # ...mismatch somewhere, include on both sides.
              inspected_properties[name] = value
            end
          end
        end

        props << {
          :node_count     => node_count,
          :digest         => resource_digest,
          :baseline_count => diffs.count,
          :accepted       => accepted,
          :can_change     => diffs.all?(&:can_change?),
          :properties     => {
            :inspected => inspected_properties,
            :baseline  => baseline_properties
          },
        }
      end
    end

    results
  end

  def self.group_against_own_baselines(group, time, start, count)
    date = time.to_date

    node_baselines = NodeBaseline.self_comparisons.all(:conditions => {:node_id => group.nodes.map(&:id), :date => date})

    group_diff(node_baselines, start, count)
  end

  def self.group_against_node(group, node, time, start, count, type=nil, title=nil)
    date = time.to_date

    node_baselines = NodeBaseline.find(:all, :conditions => {:node_id => group.nodes.map(&:id), :date => date, :reference_node_id => node.id})

    group_diff(node_baselines, start, count, type, title)
  end

end
